/*
 * This software is licensed under the terms of the MIT License.
 * See COPYING for further information.
 * ---
 * Copyright (c) 2011-2019, Lukas Weber <laochailan@web.de>.
 * Copyright (c) 2012-2019, Andrei Alexeyev <akari@taisei-project.org>.
 */

#include "taisei.h"

#include "stageutils.h"
#include "global.h"
#include "util/glm.h"
#include "video.h"

Stage3D stage_3d_context;

void stage3d_init(Stage3D *s, uint pos_buffer_size) {
	memset(s, 0, sizeof(*s));
	s->pos_buffer_size = pos_buffer_size;
	s->pos_buffer = calloc(s->pos_buffer_size, sizeof(vec3));
}

void camera3d_update(Camera3D *cam) {
	glm_vec3_add(cam->pos, cam->vel, cam->pos);
}

void stage3d_update(Stage3D *s) {
	camera3d_update(&s->cam);
}

void camera3d_apply_transforms(Camera3D *cam, mat4 mat) {
	glm_rotate_x(mat, glm_rad(-cam->rot.pitch), mat);
	glm_rotate_y(mat, glm_rad(-cam->rot.yaw), mat);
	glm_rotate_z(mat, glm_rad(-cam->rot.roll), mat);

	vec3 trans;
	glm_vec3_negate_to(cam->pos, trans);
	glm_translate(mat, trans);
}

void stage3d_apply_transforms(Stage3D *s, mat4 mat) {
	camera3d_apply_transforms(&s->cam, mat);
}

void stage3d_draw_segment(Stage3D *s, SegmentPositionRule pos_rule, SegmentDrawRule draw_rule, float maxrange) {
	uint num = pos_rule(s, s->cam.pos, maxrange);

	for(uint j = 0; j < num; ++j) {
		draw_rule(s->pos_buffer[j]);
	}
}

void stage3d_draw(Stage3D *s, float maxrange, uint nsegments, const Stage3DSegment segments[nsegments]) {
	r_mat_mv_push();
	stage3d_apply_transforms(s, *r_mat_mv_current_ptr());

	for(uint i = 0; i < nsegments; ++i) {
		const Stage3DSegment *seg = segments + i;
		stage3d_draw_segment(s, seg->pos, seg->draw, maxrange);
	}

	r_mat_mv_pop();
}

void stage3d_shutdown(Stage3D *s) {
	free(s->pos_buffer);
}


// this function generates an array of positions on a line so that all points of
// at least distance maxrange are contained.
//
// The equation of the (infinite) line is
//
//     x = support + direction * n
//
// where n is an integer.
//
uint linear3dpos(Stage3D *s3d, vec3 camera, float maxrange, vec3 support, vec3 direction) {
	vec3 support_to_camera;
	glm_vec3_sub(camera, support, support_to_camera);

	const float direction_length2 = glm_vec3_norm2(direction);
	const float projected_cam = glm_vec3_dot(support_to_camera, direction) / direction_length2;
	const int n_closest_to_cam = projected_cam;

	uint size = 0;

	// This is an approximation that does not take into account the distance
	// of the camera to the line. Can be made exact though.
	const int nrange = maxrange/sqrt(direction_length2);

	// draw nearest to closest
	for(int r = 0; r <= nrange; r++) {
		for(int dir = -1; dir <= 1; dir += 2) {
			if(r == 0 && dir > 0) {
				continue;
			}
			int n = n_closest_to_cam + dir*r;
			vec3 extended_direction;
			glm_vec3_scale(direction, n, extended_direction);

			assert(size < s3d->pos_buffer_size);
			glm_vec3_add(support, extended_direction, s3d->pos_buffer[size]);
			++size;

			if(size == s3d->pos_buffer_size) {
				s3d->pos_buffer_size *= 2;
				log_debug("pos_buffer exhausted, reallocating %u -> %u", size, s3d->pos_buffer_size);
				s3d->pos_buffer = realloc(s3d->pos_buffer, sizeof(vec3) * s3d->pos_buffer_size);
			}
		}
	}

	return size;
}

uint single3dpos(Stage3D *s3d, vec3 q, float maxrange, vec3 p) {
	assume(s3d->pos_buffer_size > 0);

	vec3 d;
	glm_vec3_sub(p, q, d);

	if(glm_vec3_norm2(d) > maxrange * maxrange) {
		return 0;
	} else {
		memcpy(s3d->pos_buffer, p, sizeof(vec3));
		return 1;
	}
}

void skip_background_anim(void (*update_func)(void), int frames, int *timer, int *timer2) {
	int targetframes = *timer + frames;

	while(*timer < targetframes) {
		++*timer;

		if(timer2) {
			++*timer2;
		}

		update_func();
	}
}
